"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var unique_stack_1 = require("./unique-stack");
var call = function (fn) { return fn(); };
var createInstanceFactory = function (factories, instances, stack, initializers) {
    if (stack === void 0) { stack = unique_stack_1.default(); }
    if (initializers === void 0) { initializers = []; }
    return function (container, name) {
        if (instances.has(name))
            return instances.get(name);
        if (stack.push(name)[0] != null) {
            throw new Error('Cyclic dependencies couldn\'t be resolved.');
        }
        var itemFactory = factories[name];
        var _a = (Array.isArray(itemFactory) ? itemFactory : [itemFactory, null]), factory = _a[0], initializer = _a[1];
        if (typeof factory !== 'function') {
            throw new Error("Factory is not defined for name \"" + name + "\"");
        }
        var instance = factory(container);
        instances.set(name, instance);
        if (typeof initializer === 'function') {
            initializers.push(function () { return initializer(instance, container, name); });
        }
        if (stack.pop(name)[0] != null) {
            throw new Error('Not all dependencies resolved correctly.');
        }
        if (stack.size === 0) {
            initializers.forEach(call);
            initializers = [];
        }
        return instance;
    };
};
exports.default = createInstanceFactory;
