"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.factoriesFrom = exports.releaseAll = exports.prepareAll = exports.isReady = void 0;
var create_instance_1 = require("./create-instance");
var all_names_1 = require("./utils/all-names");
var assert_exists_1 = require("./utils/assert-exists");
var map_object_1 = require("./utils/map-object");
var $INSTANCES = Symbol('TRUE-DI::INSTANCES');
var itemsOf = function (container) {
    return assert_exists_1.default(container[$INSTANCES], 'Argument is not a container');
};
exports.isReady = function (container, name) {
    return itemsOf(container).has(name);
};
exports.prepareAll = function (container) {
    return map_object_1.default(container, function (name) { return container[name]; });
};
exports.releaseAll = function (container) {
    itemsOf(container).clear();
};
exports.factoriesFrom = function (container, names) {
    return map_object_1.default(container, function (name) { return function () { return container[name]; }; }, names);
};
function diContainer(factories) {
    var _a;
    var instances = new Map();
    var createInstance = create_instance_1.default(factories, instances);
    var container = all_names_1.default(factories).reduce(function (containerObj, name) {
        return Object.defineProperty(containerObj, name, {
            configurable: false,
            enumerable: Object.getOwnPropertyDescriptor(factories, name).enumerable,
            get: function () { return createInstance(container, name); },
            set: function (value) {
                if (value != null) {
                    throw new Error('Container does\'t allow to replace items');
                }
                instances.delete(name);
            },
        });
    }, Object.create(Object.create(null, (_a = {},
        _a[$INSTANCES] = {
            configurable: false, writable: false, enumerable: false, value: instances,
        },
        _a))));
    return container;
}
exports.default = diContainer;
